class UDSException(Exception):
    pass


class MessageException(UDSException):
    pass


class ArgumentException(UDSException):
    pass


class RetryException(UDSException):
    pass


class InvalidVersionException(UDSException):
    link: str

    def __init__(self, client_link: str, required_version: str) -> None:
        super().__init__(client_link)
        self.link = client_link
        self.required_version = required_version



# -*- coding: utf-8 -*-
#
# Copyright (c) 2023 Virtual Cable S.L.U.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without modification,
# are permitted provided that the following conditions are met:
#
#    * Redistributions of source code must retain the above copyright notice,
#      this list of conditions and the following disclaimer.
#    * Redistributions in binary form must reproduce the above copyright notice,
#      this list of conditions and the following disclaimer in the documentation
#      and/or other materials provided with the distribution.
#    * Neither the name of Virtual Cable S.L. nor the names of its contributors
#      may be used to endorse or promote products derived from this software
#      without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

'''
Author: Adolfo Gómez, dkmaster at dkmon dot com
'''
import typing
import os
import os.path
import tempfile
import logging
import sys

from . import types


# Feature enabler
def _feature_requested(env_var: str) -> bool:
    env_var_name = env_var.upper().replace('-', '_')
    if env_var_name not in os.environ:
        # Look for temp file with that name, if it exists, its true
        # or if a file in the home directory with that name exists
        if os.path.exists(os.path.join(tempfile.gettempdir(), env_var)) or os.path.exists(
            os.path.join(os.path.expanduser('~'), env_var)
        ):
            return True

    return os.getenv(env_var_name, 'false').lower() in ('true', 'yes', '1')


DEBUG: typing.Final[bool] = _feature_requested('uds-debug-on')
LOGLEVEL: typing.Final[int] = logging.DEBUG if DEBUG else logging.INFO

LOGFILE: typing.Final[str] = os.getenv(
    'UDS_LOG_FILE',
    (
        os.path.expanduser('~/udsclient.log')  # Linux or Mac on home folder
        if 'linux' in sys.platform or 'darwin' in sys.platform
        else os.path.join(tempfile.gettempdir(), 'udsclient.log')  # Windows or unknown on temp folder
    ),
)

# UDS Client version
VERSION: typing.Final[str] = '4.0.0'

# User agent
USER_AGENT: typing.Final[str] = f'UDSClient/{VERSION} ({types.OsType.DETECTED_SO})'

# Secure channel ciphers
SECURE_CIPHERS: typing.Final[str] = (
    'TLS_AES_256_GCM_SHA384'
    ':TLS_CHACHA20_POLY1305_SHA256'
    ':TLS_AES_128_GCM_SHA256'
    ':ECDHE-RSA-AES256-GCM-SHA384'
    ':ECDHE-RSA-AES128-GCM-SHA256'
    ':ECDHE-RSA-CHACHA20-POLY1305'
    ':ECDHE-ECDSA-AES128-GCM-SHA256'
    ':ECDHE-ECDSA-AES256-GCM-SHA384'
    ':ECDHE-ECDSA-CHACHA20-POLY1305'
)

# Public key for validating signed scripts
PUBLIC_KEY = b'''-----BEGIN PUBLIC KEY-----
MIICIjANBgkqhkiG9w0BAQEFAAOCAg8AMIICCgKCAgEAuNURlGjBpqbglkTTg2lh
dU5qPbg9Q+RofoDDucGfrbY0pjB9ULgWXUetUWDZhFG241tNeKw+aYFTEorK5P+g
ud7h9KfyJ6huhzln9eyDu3k+kjKUIB1PLtA3lZLZnBx7nmrHRody1u5lRaLVplsb
FmcnptwYD+3jtJ2eK9ih935DYAkYS4vJFi2FO+npUQdYBZHPG/KwXLjP4oGOuZp0
pCTLiCXWGjqh2GWsTECby2upGS/ZNZ1r4Ymp4V2A6DZnN0C0xenHIY34FWYahbXF
ZGdr4DFBPdYde5Rb5aVKJQc/pWK0CV7LK6Krx0/PFc7OGg7ItdEuC7GSfPNV/ANt
5BEQNF5w2nUUsyN8ziOrNih+z6fWQujAAUZfpCCeV9ekbwXGhbRtdNkbAryE5vH6
eCE0iZ+cFsk72VScwLRiOhGNelMQ7mIMotNck3a0P15eaGJVE2JV0M/ag/Cnk0Lp
wI1uJQRAVqz9ZAwvF2SxM45vnrBn6TqqxbKnHCeiwstLDYG4fIhBwFxP3iMH9EqV
2+QXqdJW/wLenFjmXfxrjTRr+z9aYMIdtIkSpADIlbaJyTtuQpEdWnrlDS2b1IGd
Okbm65EebVzOxfje+8dRq9Uqwip8f/qmzFsIIsx3wPSvkKawFwb0G5h2HX5oJrk0
nVgtClKcDDlSaBsO875WDR0CAwEAAQ==
-----END PUBLIC KEY-----'''


# Variables for tunnel
BUFFER_SIZE: typing.Final[int] = 1024 * 16  # Max buffer length
LISTEN_ADDRESS: typing.Final[str] = '127.0.0.1'
LISTEN_ADDRESS_V6: typing.Final[str] = '::1'
RESPONSE_OK: typing.Final[bytes] = b'OK'

# Ticket length
TICKET_LENGTH: typing.Final[int] = 40  # Ticket length FOR UDS

# Constants strings for protocol
HANDSHAKE_V1: typing.Final[bytes] = b'\x5AMGB\xA5\x01\x00'
CMD_TEST: typing.Final[bytes] = b'TEST'
CMD_OPEN: typing.Final[bytes] = b'OPEN'



# Form implementation generated from reading ui file 'UDSWindow.ui'
#
# Created by: PySide6 UI code generator 6.5.0
#
# WARNING: Any manual changes made to this file will be lost when pyuic6 is
# run again.  Do not edit this file unless you know what you are doing.


from PySide6 import QtCore, QtGui, QtWidgets


class Ui_MainWindow(object):
    def setupUi(self, MainWindow):
        MainWindow.setObjectName("MainWindow")
        MainWindow.setWindowModality(QtCore.Qt.WindowModality.NonModal)
        MainWindow.resize(259, 185)
        MainWindow.setCursor(QtGui.QCursor(QtCore.Qt.CursorShape.BusyCursor))
        icon = QtGui.QIcon()
        icon.addPixmap(QtGui.QPixmap(":/images/logo-uds-small"), QtGui.QIcon.Mode.Normal, QtGui.QIcon.State.Off)
        MainWindow.setWindowIcon(icon)
        MainWindow.setWindowOpacity(1.0)
        self.centralwidget = QtWidgets.QWidget(parent=MainWindow)
        self.centralwidget.setAutoFillBackground(True)
        self.centralwidget.setObjectName("centralwidget")
        self.verticalLayout_2 = QtWidgets.QVBoxLayout(self.centralwidget)
        self.verticalLayout_2.setContentsMargins(4, 4, 4, 4)
        self.verticalLayout_2.setSpacing(4)
        self.verticalLayout_2.setObjectName("verticalLayout_2")
        self.frame = QtWidgets.QFrame(parent=self.centralwidget)
        self.frame.setFrameShape(QtWidgets.QFrame.Shape.StyledPanel)
        self.frame.setFrameShadow(QtWidgets.QFrame.Shadow.Raised)
        self.frame.setObjectName("frame")
        self.verticalLayout_3 = QtWidgets.QVBoxLayout(self.frame)
        self.verticalLayout_3.setContentsMargins(4, 4, 4, 4)
        self.verticalLayout_3.setSpacing(4)
        self.verticalLayout_3.setObjectName("verticalLayout_3")
        self.verticalLayout = QtWidgets.QVBoxLayout()
        self.verticalLayout.setObjectName("verticalLayout")
        self.image = QtWidgets.QLabel(parent=self.frame)
        self.image.setMinimumSize(QtCore.QSize(0, 24))
        self.image.setAutoFillBackground(True)
        self.image.setText("")
        self.image.setPixmap(QtGui.QPixmap(":/images/logo-uds-small"))
        self.image.setScaledContents(False)
        self.image.setAlignment(QtCore.Qt.AlignmentFlag.AlignCenter)
        self.image.setObjectName("image")
        self.verticalLayout.addWidget(self.image)
        self.info = QtWidgets.QLabel(parent=self.frame)
        self.info.setMaximumSize(QtCore.QSize(16777215, 16))
        self.info.setObjectName("info")
        self.verticalLayout.addWidget(self.info)
        self.progressBar = QtWidgets.QProgressBar(parent=self.frame)
        self.progressBar.setProperty("value", 24)
        self.progressBar.setTextVisible(False)
        self.progressBar.setObjectName("progressBar")
        self.verticalLayout.addWidget(self.progressBar)
        self.horizontalLayout = QtWidgets.QHBoxLayout()
        self.horizontalLayout.setObjectName("horizontalLayout")
        spacerItem = QtWidgets.QSpacerItem(40, 20, QtWidgets.QSizePolicy.Policy.Expanding, QtWidgets.QSizePolicy.Policy.Minimum)
        self.horizontalLayout.addItem(spacerItem)
        self.cancelButton = QtWidgets.QPushButton(parent=self.frame)
        self.cancelButton.setDefault(True)
        self.cancelButton.setFlat(False)
        self.cancelButton.setObjectName("cancelButton")
        self.horizontalLayout.addWidget(self.cancelButton)
        spacerItem1 = QtWidgets.QSpacerItem(40, 20, QtWidgets.QSizePolicy.Policy.Expanding, QtWidgets.QSizePolicy.Policy.Minimum)
        self.horizontalLayout.addItem(spacerItem1)
        self.verticalLayout.addLayout(self.horizontalLayout)
        self.verticalLayout_3.addLayout(self.verticalLayout)
        self.verticalLayout_2.addWidget(self.frame)
        MainWindow.setCentralWidget(self.centralwidget)

        self.retranslateUi(MainWindow)
        QtCore.QMetaObject.connectSlotsByName(MainWindow)

    def retranslateUi(self, MainWindow):
        _translate = QtCore.QCoreApplication.translate
        MainWindow.setWindowTitle(_translate("MainWindow", "UDS Connection"))
        self.info.setText(_translate("MainWindow", "TextLabel"))
        self.cancelButton.setText(_translate("MainWindow", "Cancel"))



# -*- coding: utf-8 -*-
#
# Copyright (c) 2017-2024 Virtual Cable S.L.U.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without modification,
# are permitted provided that the following conditions are met:
#
#    * Redistributions of source code must retain the above copyright notice,
#      this list of conditions and the following disclaimer.
#    * Redistributions in binary form must reproduce the above copyright notice,
#      this list of conditions and the following disclaimer in the documentation
#      and/or other materials provided with the distribution.
#    * Neither the name of Virtual Cable S.L. nor the names of its contributors
#      may be used to endorse or promote products derived from this software
#      without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

'''
Author: Adolfo Gómez, dkmaster at dkmon dot com
'''
import base64
import bz2
import json
import socket
import ssl
import typing
import urllib
import urllib.error
import urllib.parse
import urllib.request

from cryptography import x509
from cryptography.hazmat.backends import default_backend

from . import consts, tools, exceptions
from .log import logger

# Callback for error on cert
# parameters are hostname, serial
# If returns True, ignores error
CertCallbackType = typing.Callable[[str, str], bool]


class RestApi:

    _rest_api_endpoint: str  # base Rest API URL
    _on_invalid_certificate: typing.Optional[CertCallbackType]
    _required_version: str

    def __init__(
        self,
        rest_api_endpoint: str,
        on_invalid_certificate: typing.Optional[CertCallbackType] = None,
    ) -> None:  # parent not used
        logger.debug('Setting request URL to %s', rest_api_endpoint)

        self._rest_api_endpoint = rest_api_endpoint
        self._on_invalid_certificate = on_invalid_certificate
        self._required_version = ''

    def request(
        self,
        path: str,
        params: typing.Optional[typing.Mapping[str, str]] = None,
        data: typing.Optional[str] = None,
    ) -> typing.Any:
        if params:
            path += '?' + '&'.join(
                '{}={}'.format(k, urllib.parse.quote(str(v).encode('utf8'))) for k, v in params.items()
            )

        return json.loads(RestApi.request_url(self._rest_api_endpoint + path, self._on_invalid_certificate, data=data))

    def process_error(self, data: typing.Any) -> None:
        if 'error' in data:
            # Get retrayable from data, if not present, use old key
            if data.get('is_retrayable', data.get('retryable', '0')) == '1':
                raise exceptions.RetryException(data['error'])

            raise exceptions.UDSException(data['error'])

    def get_version(self) -> str:
        '''Gets and stores the serverVersion.
        Also checks that the version is valid for us. If not,
        will raise an "InvalidVersion' exception'''

        client_link = ''
        if not self._required_version:
            data = self.request('')  # Version is returned on 'main' path
            self.process_error(data)
            # get server version, using new key but, if not present, use old one
            # Note: old version will be removed on 5.0.0 (As all 4.0 brokers will already return the new keys)
            if 'requiredVersion' in data['result']:
                self._required_version = data['result']['requiredVersion']
                client_link = data['result']['downloadUrl']
            else:
                self._required_version = data['result']['required_version']
                client_link = data['result']['client_link']

        try:
            if self._required_version > consts.VERSION:
                raise exceptions.InvalidVersionException(client_link, self._required_version)

            return self._required_version
        except exceptions.InvalidVersionException:
            raise
        except Exception as e:
            raise exceptions.UDSException(e) from e

    def get_script_and_parameters(
        self, ticket: str, scrambler: str
    ) -> typing.Tuple[str, typing.Any, typing.Dict[str, typing.Any]]:
        '''Gets the transport script, validates it if necesary
        and returns it'''
        try:
            data = self.request(
                '/{}/{}'.format(ticket, scrambler),
                params={'hostname': tools.gethostname(), 'version': consts.VERSION},
            )
        except Exception as e:
            logger.exception('Got exception on getTransportData')
            raise e

        logger.debug('Transport data received')
        self.process_error(data)

        params = None

        res: typing.Dict[str, typing.Any] = data['result']
        # We have three elements on result:
        # * Script
        # * Signature
        # * Script data
        # We test that the Script has correct signature, and them execute it with the parameters
        # script, signature, params = res['script'].decode('base64').decode('bz2'), res['signature'], json.loads(res['params'].decode('base64').decode('bz2'))
        script, signature, params = (
            bz2.decompress(base64.b64decode(res['script'])),
            res['signature'],
            json.loads(bz2.decompress(base64.b64decode(res['params']))),
        )
        if tools.verify_signature(script, signature) is False:
            logger.error('Signature is invalid')

            raise Exception('Invalid UDS code signature. Please, report to administrator')

        return script.decode(), params, res.get('log', {})

        # exec(script.decode("utf-8"), globals(), {'parent': self, 'sp': params})

    def send_log(self, log_ticket: str, log_data: str) -> None:
        '''Sends log data to server'''
        try:
            data = self.request(
                '/{}/log'.format(log_ticket),
                data=json.dumps({'log': log_data}),
                params={'hostname': tools.gethostname(), 'version': consts.VERSION},
            )
        except Exception as e:
            logger.exception('Got exception on sendLog')
            raise e

        self.process_error(data)

    @staticmethod
    def _open(
        url: str, certErrorCallback: typing.Optional[CertCallbackType] = None, data: 'bytes|None' = None
    ) -> typing.Any:
        ctx = ssl.create_default_context()
        ctx.check_hostname = False
        ctx.verify_mode = ssl.CERT_NONE
        # Disable SSLv2, SSLv3, TLSv1, TLSv1.1
        ctx.minimum_version = ssl.TLSVersion.TLSv1_2
        ctx.set_ciphers(consts.SECURE_CIPHERS)

        # If we have the certificates file, we use it
        if tools.get_cacerts_file() is not None:
            ctx.load_verify_locations(tools.get_cacerts_file())
        hostname = urllib.parse.urlparse(url)[1]
        serial = ''

        port = ''
        if ':' in hostname:
            hostname, port = hostname.split(':')

        if url.startswith('https'):
            port = port or '443'
            with ctx.wrap_socket(
                socket.socket(socket.AF_INET, socket.SOCK_STREAM),
                server_hostname=hostname,
            ) as s:
                s.connect((hostname, int(port)))
                # Get binary certificate
                binCert = s.getpeercert(True)
                if binCert:
                    cert = x509.load_der_x509_certificate(binCert, default_backend())
                else:
                    raise Exception('Certificate not found!')

            serial = hex(cert.serial_number)[2:]

        response = None
        ctx.verify_mode = ssl.CERT_REQUIRED
        ctx.check_hostname = True

        def _open_url(url: str) -> typing.Any:
            # Generate the request with the headers
            req = urllib.request.Request(
                url,
                headers={
                    'User-Agent': consts.USER_AGENT,
                },
            )
            return urllib.request.urlopen(req, data, context=ctx)

        try:
            response = _open_url(url)
        except urllib.error.URLError as e:
            if isinstance(e.reason, ssl.SSLCertVerificationError):
                # Ask about invalid certificate
                if certErrorCallback:
                    if certErrorCallback(hostname, serial):
                        ctx.check_hostname = False
                        ctx.verify_mode = ssl.CERT_NONE
                        response = _open_url(url)
                else:
                    raise
            else:
                raise

        return response

    @staticmethod
    def api(host: str, on_invalid_certificate: CertCallbackType) -> 'RestApi':
        return RestApi(f'https://{host}/uds/rest/client', on_invalid_certificate)

    @staticmethod
    def request_url(
        url: str, cert_error_callback: typing.Optional[CertCallbackType] = None, data: 'str|None' = None
    ) -> bytes:
        with RestApi._open(url, cert_error_callback, data=data.encode() if data else None) as response:
            resp = response.read()

        return resp



# -*- coding: utf-8 -*-
#
# Copyright (c) 2023 Virtual Cable S.L.U.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without modification,
# are permitted provided that the following conditions are met:
#
#    * Redistributions of source code must retain the above copyright notice,
#      this list of conditions and the following disclaimer.
#    * Redistributions in binary form must reproduce the above copyright notice,
#      this list of conditions and the following disclaimer in the documentation
#      and/or other materials provided with the distribution.
#    * Neither the name of Virtual Cable S.L.U. nor the names of its contributors
#      may be used to endorse or promote products derived from this software
#      without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

'''
Author: Adolfo Gómez, dkmaster at dkmon dot com
'''
import logging
import socket
import base64

import urllib.request

# This module allows to use proxyes to make socket connections
# It's simply initializes the connection and, after proxy (if required) is connected, returns the socket
# Currently experimenting with this, but not used at all


def connect(host: str, port: int) -> socket.socket:
    # Get proxy settings
    proxy = urllib.request.getproxies().get('https', None)
    if proxy:
        try:
            # If proxy is set, connect to it and try to connect to host:port
            # First, extract proxy scheme to connect
            # proxy_scheme = proxy[: proxy.index('://')]
            # Remove scheme from proxy
            proxy = proxy[proxy.index('://') + 3 :]

            # If user:password is present in proxy, it will be used
            if '@' in proxy:
                # Extract user:password
                proxy_user_password = proxy[: proxy.index('@')]
                # Remove user:password from proxy
                proxy = proxy[proxy.index('@') + 1 :]
                # Encode user and password for basic auth on proxy
                proxy_user_password = base64.b64encode(proxy_user_password.encode('utf8')).decode('utf8')
            else:
                proxy_user_password = None

            # ProxyHost may be ipv4 or ipv6, so we need to split it
            proxy_host, proxy_port_str = proxy.rsplit(':', 1)
            proxy_port = int(proxy_port_str)
            if proxy_host.startswith('['):
                # ipv6
                proxy_host = proxy_host[1:-1]
            logging.debug(
                'Connecting to proxy {}:{} to connect to {}:{}'.format(proxy_host, proxy_port, host, port)
            )
            s = socket.socket(socket.AF_INET6 if ':' in proxy_host else socket.AF_INET, socket.SOCK_STREAM)
            s.connect((proxy_host, proxy_port))
            # if https proxy, we need to upgrade connection to https
            s.sendall(f'CONNECT {host}:{port} HTTP/1.1\r\n'.encode('utf8'))
            if proxy_user_password:
                s.sendall(f'Proxy-Authorization: Basic {proxy_user_password}\r\n'.encode('utf8'))
            s.sendall(b'\r\n')
            # Read response
            data = s.recv(4096)
            if not data.startswith(b'HTTP/1.1 200'):
                raise Exception(f'Proxy returned error: {data!r}')
            # Return socket
            return s
        except Exception as e:
            logging.error('Error connecting to proxy: %s. Trying direct connection', e)
            # fall back to direct connection

    # If no proxy is set, simply connect to host:port
    logging.debug('Connecting to {}:{}'.format(host, port))
    s = socket.socket(socket.AF_INET6 if ':' in host else socket.AF_INET, socket.SOCK_STREAM)
    s.connect((host, port))
    return s


if __name__ == "__main__":
    import os

    os.environ['https_proxy'] = 'http://proxy:3128'
    logging.basicConfig(level=logging.DEBUG)
    s = connect('www.google.com', 80)
    s.sendall(b'GET / HTTP/1.0\r\nHost: www.google.com\r\n\r\n')
    response = b''
    while True:
        data = s.recv(1024)
        if not data:
            break
        response += data
    s.close()
    print(response.decode('utf8'))



import sys
import os.path
import subprocess
import typing

from uds.log import logger
import UDSClient

# Basically, this will be the main file for the Mac OS X application
# It will be called from the Info.plist file, and will be responsible for
# launching the UDSClient application, and also for handling the URL
# that will be passed to it.
# In order to do so, we will call "ourselves" with the URL as parameter, and
# in that case, we will process the URL and launch the UDSClient application
# If no url is passed, UDSClientLauncher will be instantiated, waiting for
# the event that will be generated when the application is called with the
# URL as parameter.

from uds.ui import QtCore, QtWidgets, QtGui, Ui_MacLauncher

SCRIPT_NAME = 'UDSClientLauncher'

class UDSClientLauncher(QtWidgets.QApplication):
    path: str
    tunnels: 'typing.List[subprocess.Popen[typing.Any]]'

    def __init__(self, argv: typing.List[str]) -> None:
        super().__init__(argv)
        self.path = os.path.join(os.path.dirname(sys.argv[0]).replace('Resources', 'MacOS'), SCRIPT_NAME)
        self.tunnels = []
        self.lastWindowClosed.connect(self.close_tunnels)

    def clean_tunnels(self) -> None:
        '''
        Removes all finished tunnels from the list
        '''

        def _is_running(p: subprocess.Popen[typing.Any]) -> bool:
            try:
                if p.poll() is None:
                    return True
            except Exception as e:
                logger.debug('Got error polling subprocess: %s', e)
            return False

        # Remove references to finished tunnels, they will be garbage collected
        self.tunnels = [tunnel for tunnel in self.tunnels if _is_running(tunnel)]

    def close_tunnels(self) -> None:
        '''
        Finishes all running tunnels
        '''
        logger.debug('Closing remaining tunnels')
        for tunnel in self.tunnels:
            logger.debug('Checking %s - "%s"', tunnel, tunnel.poll())
            if tunnel.poll() is None:  # Running
                logger.info('Found running tunnel %s, closing it', tunnel.pid)
                tunnel.kill()

    def event(self, evnt: QtCore.QEvent) -> bool:  # pyright: ignore[reportIncompatibleMethodOverride]
        if evnt.type() == QtCore.QEvent.Type.FileOpen:
            fe = typing.cast(QtGui.QFileOpenEvent, evnt)
            logger.debug('Got url: %s', fe.url().url())
            fe.accept()
            logger.debug('Spawning %s', self.path)
            # First, remove all finished tunnel processed from check queue, to keelp things clean
            self.clean_tunnels()
            # And now add a new one, calling self with the url
            self.tunnels.append(subprocess.Popen([self.path, fe.url().url()]))

        return super().event(evnt)


def main(args: typing.List[str]) -> None:
    if len(args) > 1:
        UDSClient.main(args)
    else:
        app = UDSClientLauncher([])  # no args for launcher needed
        window = QtWidgets.QMainWindow()
        Ui_MacLauncher().setupUi(window)  # type: ignore

        window.showMinimized()

        sys.exit(app.exec())


if __name__ == "__main__":
    main(args=sys.argv)



# -*- coding: utf-8 -*-
#
# Copyright (c) 2022 Virtual Cable S.L.U.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without modification,
# are permitted provided that the following conditions are met:
#
#    * Redistributions of source code must retain the above copyright notice,
#      this list of conditions and the following disclaimer.
#    * Redistributions in binary form must reproduce the above copyright notice,
#      this list of conditions and the following disclaimer in the documentation
#      and/or other materials provided with the distribution.
#    * Neither the name of Virtual Cable S.L. nor the names of its contributors
#      may be used to endorse or promote products derived from this software
#      without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
'''
@author: Adolfo Gómez, dkmaster at dkmon dot com
'''
import contextlib
import socket
import socketserver
import ssl
import threading
import select
import typing
import logging

from . import tools, consts, types


logger = logging.getLogger(__name__)


class ForwardServer(socketserver.ThreadingTCPServer):
    daemon_threads = True
    allow_reuse_address = True

    remote: typing.Tuple[str, int]
    remote_ipv6: bool
    ticket: str
    stop_flag: threading.Event
    can_stop: bool
    timer: typing.Optional[threading.Timer]
    check_certificate: bool
    keep_listening: bool
    current_connections: int
    status: types.ForwardState

    address_family = socket.AF_INET

    def __init__(
        self,
        remote: typing.Tuple[str, int],
        ticket: str,
        timeout: int = 0,
        local_port: int = 0,  # Use first available listen port if not specified
        check_certificate: bool = True,
        keep_listening: bool = False,
        ipv6_listen: bool = False,
        ipv6_remote: bool = False,
    ) -> None:
        # Negative values for timeout, means:
        #   * accept always connections, but if no connection is stablished on timeout
        #     (positive), stop the listener
        #
        # Note that this is for backwards compatibility, better use "keep_listening"
        if timeout < 0:
            keep_listening = True
            timeout = abs(timeout)

        if ipv6_listen:
            self.address_family = socket.AF_INET6

        # Binds and activate the server, so if local_port is 0, it will be assigned
        super().__init__(
            server_address=(consts.LISTEN_ADDRESS_V6 if ipv6_listen else consts.LISTEN_ADDRESS, local_port),
            RequestHandlerClass=Handler,
        )

        self.remote = remote
        self.remote_ipv6 = ipv6_remote or ':' in remote[0]  # if ':' in remote address, it's ipv6 (port is [1])
        self.ticket = ticket
        self.check_certificate = check_certificate
        self.keep_listening = keep_listening
        self.stop_flag = threading.Event()  # False initial
        self.current_connections = 0

        self.status = types.ForwardState.TUNNEL_LISTENING
        self.can_stop = False

        timeout = timeout or 60
        self.timer = threading.Timer(timeout, ForwardServer._set_stoppable, args=(self,))
        self.timer.start()

        logger.debug('Remote: %s', remote)
        logger.debug('Remote IPv6: %s', self.remote_ipv6)
        logger.debug('Ticket: %s', ticket)
        logger.debug('Check certificate: %s', check_certificate)
        logger.debug('Keep listening: %s', keep_listening)
        logger.debug('Timeout: %s', timeout)

    def stop(self) -> None:
        if not self.stop_flag.is_set():
            logger.debug('Stopping servers')
            self.stop_flag.set()
            if self.timer:
                self.timer.cancel()
                self.timer = None
            self.shutdown()

    @contextlib.contextmanager
    def connection(self) -> typing.Generator[ssl.SSLSocket, None, None]:
        ssl_sock: typing.Optional[ssl.SSLSocket] = None
        try:
            ssl_sock = ForwardServer._connect(self.remote, self.remote_ipv6, self.check_certificate)
            yield ssl_sock
        finally:
            if ssl_sock:
                ssl_sock.close()

    def check(self) -> bool:
        if self.status == types.ForwardState.TUNNEL_ERROR:
            return False

        logger.debug('Checking tunnel availability')

        with self.connection() as ssl_socket:
            return ForwardServer._test(ssl_socket)

    @contextlib.contextmanager
    def open_tunnel(self) -> typing.Generator[ssl.SSLSocket, None, None]:
        logger.debug('Tunnel opening')
        self.current_connections += 1
        logger.debug('Opening tunnel with ticket %s, connections: %s', self.ticket, self.current_connections)
        # Open remote connection
        try:
            with self.connection() as ssl_socket:
                ForwardServer._open_tunnel(ssl_socket, self.ticket)

                yield ssl_socket
        except ssl.SSLError as e:
            logger.error(f'Certificate error connecting to {self.remote!s}: {e!s}')
            self.status = types.ForwardState.TUNNEL_ERROR
            self.stop()
        except Exception as e:
            logger.error(f'Error connecting to {self.remote!s}: {e!s}')
            self.status = types.ForwardState.TUNNEL_ERROR
            self.stop()
        finally:
            logger.debug('Decrementing connections to %s', self.current_connections-1)
            self.current_connections -= 1

    @property
    def stoppable(self) -> bool:
        logger.debug('Is stoppable: %s', self.can_stop)
        return self.can_stop

    @staticmethod
    def _set_stoppable(fs: 'ForwardServer') -> None:
        # As soon as the timer is fired, the server can be stopped
        # This means that:
        #  * If not connections are stablished, the server will be stopped
        #  * If no "keep_listening" is set, the server will not allow any new connections
        logger.debug('New connection limit reached')
        fs.timer = None
        fs.can_stop = True
        # If timer fired, and no connections are stablished, stop the server
        logger.debug('Current connections: %s', fs.current_connections)
        if fs.current_connections <= 0:
            fs.stop()

    @staticmethod
    def _test(ssl_socket: ssl.SSLSocket) -> bool:
        try:
            ssl_socket.sendall(consts.CMD_TEST)
            resp = ssl_socket.recv(2)
            if resp != consts.RESPONSE_OK:
                raise Exception({'Invalid  tunnelresponse: {resp}'})
            logger.debug('Tunnel is available!')
            return True
        except ssl.SSLError as e:
            logger.error(f'Certificate error connecting to {ssl_socket.getsockname()}: {e!s}')
            # will surpas the "check" method on script caller, arriving to the UDSClient error handler
            raise Exception(f'Certificate error connecting to {ssl_socket.getsockname()}') from e
        except Exception as e:
            logger.error('Error connecting to tunnel server %s: %s', ssl_socket.getsockname(), e)
        return False

    @staticmethod
    def _open_tunnel(ssl_socket: ssl.SSLSocket, ticket: str) -> None:
        # Send handhshake + command + ticket
        ssl_socket.sendall(consts.CMD_OPEN + ticket.encode())
        # Check response is OK
        data = ssl_socket.recv(2)
        if data != consts.RESPONSE_OK:
            data += ssl_socket.recv(128)
            raise Exception(f'Error received: {data.decode(errors="ignore")}')  # Notify error

    @staticmethod
    def _connect(
        remote_addr: typing.Tuple[str, int],
        use_ipv6: bool = False,
        check_certificate: bool = True,
    ) -> ssl.SSLSocket:
        with socket.socket(socket.AF_INET6 if use_ipv6 else socket.AF_INET, socket.SOCK_STREAM) as rsocket:
            logger.info('CONNECT to %s', remote_addr)

            rsocket.connect(remote_addr)

            rsocket.sendall(consts.HANDSHAKE_V1)  # No response expected, just the handshake

            # Now, upgrade to ssl
            context = ssl.create_default_context()

            # Do not "recompress" data, use only "base protocol" compression
            context.options |= ssl.OP_NO_COMPRESSION
            # Macs with default installed python, does not support mininum tls version set to TLSv1.3
            # USe "brew" version instead, or uncomment next line and comment the next one
            # context.minimum_version = ssl.TLSVersion.TLSv1_2 if tools.isMac() else ssl.TLSVersion.TLSv1_3
            # Disallow old versions of TLS
            # context.minimum_version = ssl.TLSVersion.TLSv1_2
            # Secure ciphers, use this is enabled tls 1.2
            # context.set_ciphers('ECDHE-RSA-AES256-GCM-SHA512:DHE-RSA-AES256-GCM-SHA512:ECDHE-RSA-AES256-GCM-SHA384:DHE-RSA-AES256-GCM-SHA384:ECDHE-RSA-AES256-SHA384')

            context.minimum_version = ssl.TLSVersion.TLSv1_3

            if tools.get_cacerts_file() is not None:
                context.load_verify_locations(tools.get_cacerts_file())  # Load certifi certificates

            # If ignore remote certificate
            if check_certificate is False:
                context.check_hostname = False
                context.verify_mode = ssl.CERT_NONE
                logger.warning('Certificate checking is disabled!')

            return context.wrap_socket(rsocket, server_hostname=remote_addr[0])


class Handler(socketserver.BaseRequestHandler):
    # Override Base type
    server: ForwardServer  # pyright: ignore[reportIncompatibleVariableOverride]

    def handle(self) -> None:
        if self.server.status == types.ForwardState.TUNNEL_LISTENING:
            self.server.status = types.ForwardState.TUNNEL_OPENING  # Only update state on first connection

        # If server new connections processing are over time...
        if self.server.stoppable and not self.server.keep_listening:
            self.server.status = types.ForwardState.TUNNEL_ERROR
            logger.error('Rejected connection out of time')
            self.request.close()  # End connection without processing it
            return

        # Open remote connection
        self.establish_and_handle_tunnel()

        # If no more connections are stablished, and server is stoppable, do it now
        if self.server.current_connections <= 0 and self.server.stoppable:
            logger.info('No more connections, stopping server')
            self.server.stop()

    def establish_and_handle_tunnel(self) -> None:
        # Open remote connection
        try:
            with self.server.open_tunnel() as ssl_socket:
                self.handle_tunnel(remote=ssl_socket)
        except ssl.SSLError as e:
            logger.error(f'Certificate error connecting to {self.server.remote!s}: {e!s}')
            self.server.status = types.ForwardState.TUNNEL_ERROR
            self.server.stop()
        except Exception as e:
            logger.error(f'Error connecting to {self.server.remote!s}: {e!s}')
            self.server.status = types.ForwardState.TUNNEL_ERROR
            self.server.stop()

    # Processes data forwarding
    def handle_tunnel(self, remote: ssl.SSLSocket) -> None:
        self.server.status = types.ForwardState.TUNNEL_PROCESSING
        logger.debug('Start processing tunnel for ticket %s', self.server.ticket)
        # Process data until stop requested or connection closed
        try:
            readables = [self.request, remote]
            writables: typing.List[typing.Any] = []
            exceptionables = [self.request, remote]

            while not self.server.stop_flag.is_set():
                # Wait for data from either side
                readable, _writable, exceptional = select.select(readables, writables, exceptionables, 1)
                for err in exceptional:
                    raise Exception(f'Error on connection: {err}')

                if self.request in readable:  # If request (local) has data, send to remote
                    data = self.request.recv(consts.BUFFER_SIZE)
                    # End of data, close connection
                    if not data:
                        # close remote and local connections
                        logger.debug('End of local data, closing connection')
                        break
                    remote.sendall(data)
                if remote in readable:  # If remote has data, send to request (local)
                    data = remote.recv(consts.BUFFER_SIZE)
                    if not data:
                        # close local (request) and remote connections
                        logger.debug('End of remote data, closing connection')
                        break
                    self.request.sendall(data)
            logger.debug('Finished tunnel with ticket %s', self.server.ticket)
        except Exception as e:
            logger.info('Remote connection closed: %s', e)

        finally:
            try:
                self.request.close()
            except Exception as e:
                logger.error('Error closing local connection: %s', e)

            try:
                remote.close()
            except Exception as e:
                logger.error('Error closing remote connection: %s', e)


def _run(server: ForwardServer) -> None:
    """
    Runs the forwarder server.
    This method is intended to be run in a separate thread.

    Args:
        server (ForwardServer): The forward server instance.

    Returns:
        None
    """

    def _runner() -> None:
        logger.debug(
            'Starting forwarder: %s -> %s',
            server.server_address,
            server.remote,
        )
        server.serve_forever()
        logger.debug('Stopped forwarder %s -> %s', server.server_address, server.remote)

    threading.Thread(target=_runner).start()


def forward(
    remote: typing.Tuple[str, int],
    ticket: str,
    timeout: int = 0,
    local_port: int = 0,
    check_certificate: bool = True,
    keep_listening: bool = True,
    use_ipv6: bool = False,
) -> ForwardServer:
    """
    Forward a connection to a remote server.

    Args:
        remote (Tuple[str, int]): The address and port of the remote server.
        ticket (str): The ticket used for authentication.
        timeout (int, optional): When the server will stop listening for new connections (default is 0, which means never).
        local_port (int, optional): The local port to bind to (default is 0, which means any available port).
        check_certificate (bool, optional): Whether to check the server's SSL certificate (default is True).
        keep_listening (bool, optional): Whether to keep listening for new connections (default is True).

    Returns:
        ForwardServer: An instance of the ForwardServer class.

    """
    fs = ForwardServer(
        remote=remote,
        ticket=ticket,
        timeout=timeout,
        local_port=local_port,
        check_certificate=check_certificate,
        ipv6_remote=use_ipv6,
        keep_listening=keep_listening,
    )
    # Starts a new thread for processing the server,
    # so the main thread can continue processing other tasks
    _run(fs)

    return fs



# Form implementation generated from reading ui file 'UDSLauncherMac.ui'
#
# Created by: PyQt6 UI code generator 6.5.0
#
# WARNING: Any manual changes made to this file will be lost when pyuic6 is
# run again.  Do not edit this file unless you know what you are doing.


from PyQt6 import QtCore, QtGui, QtWidgets


class Ui_MacLauncher(object):
    def setupUi(self, MacLauncher):
        MacLauncher.setObjectName("MacLauncher")
        MacLauncher.setWindowModality(QtCore.Qt.WindowModality.NonModal)
        MacLauncher.resize(235, 120)
        MacLauncher.setCursor(QtGui.QCursor(QtCore.Qt.CursorShape.ArrowCursor))
        icon = QtGui.QIcon()
        icon.addPixmap(QtGui.QPixmap(":/images/logo-uds-small"), QtGui.QIcon.Mode.Normal, QtGui.QIcon.State.Off)
        MacLauncher.setWindowIcon(icon)
        MacLauncher.setWindowOpacity(1.0)
        self.centralwidget = QtWidgets.QWidget(parent=MacLauncher)
        self.centralwidget.setAutoFillBackground(True)
        self.centralwidget.setObjectName("centralwidget")
        self.verticalLayout_2 = QtWidgets.QVBoxLayout(self.centralwidget)
        self.verticalLayout_2.setContentsMargins(4, 4, 4, 4)
        self.verticalLayout_2.setSpacing(4)
        self.verticalLayout_2.setObjectName("verticalLayout_2")
        self.frame = QtWidgets.QFrame(parent=self.centralwidget)
        self.frame.setFrameShape(QtWidgets.QFrame.Shape.StyledPanel)
        self.frame.setFrameShadow(QtWidgets.QFrame.Shadow.Raised)
        self.frame.setObjectName("frame")
        self.verticalLayout = QtWidgets.QVBoxLayout(self.frame)
        self.verticalLayout.setObjectName("verticalLayout")
        self.topLabel = QtWidgets.QLabel(parent=self.frame)
        self.topLabel.setTextFormat(QtCore.Qt.TextFormat.RichText)
        self.topLabel.setObjectName("topLabel")
        self.verticalLayout.addWidget(self.topLabel)
        self.image = QtWidgets.QLabel(parent=self.frame)
        self.image.setMinimumSize(QtCore.QSize(0, 32))
        self.image.setAutoFillBackground(True)
        self.image.setText("")
        self.image.setPixmap(QtGui.QPixmap(":/images/logo-uds-small"))
        self.image.setScaledContents(False)
        self.image.setAlignment(QtCore.Qt.AlignmentFlag.AlignCenter)
        self.image.setObjectName("image")
        self.verticalLayout.addWidget(self.image)
        self.label_2 = QtWidgets.QLabel(parent=self.frame)
        self.label_2.setTextFormat(QtCore.Qt.TextFormat.RichText)
        self.label_2.setObjectName("label_2")
        self.verticalLayout.addWidget(self.label_2)
        self.verticalLayout_2.addWidget(self.frame)
        MacLauncher.setCentralWidget(self.centralwidget)

        self.retranslateUi(MacLauncher)
        QtCore.QMetaObject.connectSlotsByName(MacLauncher)

    def retranslateUi(self, MacLauncher):
        _translate = QtCore.QCoreApplication.translate
        MacLauncher.setWindowTitle(_translate("MacLauncher", "UDS Launcher"))
        self.topLabel.setText(_translate("MacLauncher", "<html><head/><body><p align=\"center\"><span style=\" font-size:12pt; font-weight:600;\">UDS Launcher</span></p></body></html>"))
        self.label_2.setText(_translate("MacLauncher", "<html><head/><body><p align=\"center\"><span style=\" font-size:6pt;\">Closing this window will end all UDS tunnels</span></p></body></html>"))



# -*- coding: utf-8 -*-
#
# Copyright (c) 2023 Virtual Cable S.L.U.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without modification,
# are permitted provided that the following conditions are met:
#
#    * Redistributions of source code must retain the above copyright notice,
#      this list of conditions and the following disclaimer.
#    * Redistributions in binary form must reproduce the above copyright notice,
#      this list of conditions and the following disclaimer in the documentation
#      and/or other materials provided with the distribution.
#    * Neither the name of Virtual Cable S.L. nor the names of its contributors
#      may be used to endorse or promote products derived from this software
#      without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

'''
Author: Adolfo Gómez, dkmaster at dkmon dot com
'''
import typing

# First, try to use PySide6, available on arm64, x86_64, i386, ...
try:
    from PySide6 import QtCore, QtWidgets, QtGui  # type: ignore
    from PySide6.QtCore import QSettings          # type: ignore

    from .pyside6.UDSLauncherMac import Ui_MacLauncher
    from .pyside6.UDSWindow import Ui_MainWindow
    from .pyside6 import UDSResources_rc
    _version = 'PySide6'
# second, try to use PyQt6, available on arm64, x86_64, i386, and also as packages for different os (debian, ubuntu, fedora, opensuse, ...)
except ImportError:
    try:
        from PyQt6 import QtCore, QtWidgets, QtGui  # type: ignore
        from PyQt6.QtCore import QSettings  # type: ignore

        from .qt6.UDSLauncherMac import Ui_MacLauncher  # type: ignore
        from .qt6.UDSWindow import Ui_MainWindow  # type: ignore
        from .qt6 import UDSResources_rc  # type: ignore
        _version = 'PyQt6'

    except ImportError:  # If not found, try to use PyQt5 (not available on arm64)
        from PyQt5 import QtCore, QtWidgets, QtGui  # type: ignore
        from PyQt5.QtCore import QSettings  # type: ignore

        from .qt5.UDSLauncherMac import Ui_MacLauncher  # type: ignore
        from .qt5.UDSWindow import Ui_MainWindow  # type: ignore
        from .qt5 import UDSResources_rc  # type: ignore
        _version = 'PyQt5'

QT_VERSION: typing.Final[str] = _version

__all__ = ['QtCore', 'QtWidgets', 'QtGui', 'Ui_MacLauncher', 'Ui_MainWindow', 'UDSResources_rc', 'QSettings', 'QT_VERSION']



# -*- coding: utf-8 -*-
#
# Copyright (c) 2015-2021 Virtual Cable S.L.U.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without modification,
# are permitted provided that the following conditions are met:
#
#    * Redistributions of source code must retain the above copyright notice,
#      this list of conditions and the following disclaimer.
#    * Redistributions in binary form must reproduce the above copyright notice,
#      this list of conditions and the following disclaimer in the documentation
#      and/or other materials provided with the distribution.
#    * Neither the name of Virtual Cable S.L. nor the names of its contributors
#      may be used to endorse or promote products derived from this software
#      without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

'''
@author: Adolfo Gómez, dkmaster at dkmon dot com
'''
import base64
import os
import os.path
import random
import socket
import stat
import string
import sys
import tempfile
import time
import typing

import certifi

# For signature checking
from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives import hashes, serialization
from cryptography.hazmat.primitives.asymmetric import padding

from uds import types

try:
    import psutil

    def process_iter(*args: typing.Any, **kwargs: typing.Any) -> typing.Any:
        return psutil.process_iter(*args, **kwargs)

except ImportError:

    def process_iter(*args: typing.Any, **kwargs: typing.Any) -> typing.Any:
        return []


from . import consts
from .log import logger


# Global variables is fine, no more than one thread will be running
# at the same time for the same process, so no need to lock
_unlink_files: typing.List[types.RemovableFile] = []
_awaitable_tasks: typing.List[types.AwaitableTask] = []
_execute_before_exit: typing.List[typing.Callable[[], None]] = []

sys_fs_enc = sys.getfilesystemencoding() or 'mbcs'


def save_temp_file(content: str, filename: typing.Optional[str] = None) -> str:
    if filename is None:
        filename = ''.join(random.choice(string.ascii_lowercase + string.digits) for _ in range(16))
        filename = filename + '.uds'

    filename = os.path.join(tempfile.gettempdir(), filename)

    with open(filename, 'w') as f:
        f.write(content)

    logger.info('Returning filename')
    return filename


def read_temp_file(filename: str) -> typing.Optional[str]:
    filename = os.path.join(tempfile.gettempdir(), filename)
    try:
        with open(filename, 'r') as f:
            return f.read()
    except Exception:
        return None


def test_server(host: str, port: typing.Union[str, int], timeout: int = 4) -> bool:
    try:
        sock = socket.create_connection((host, int(port)), timeout)
        sock.close()
    except Exception:
        return False
    return True


def find_application(application_name: str, extra_path: typing.Optional[str] = None) -> typing.Optional[str]:
    searchPath = os.environ['PATH'].split(os.pathsep)
    if extra_path:
        searchPath += list(extra_path)

    for path in searchPath:
        fileName = os.path.join(path, application_name)
        if os.path.isfile(fileName) and (os.stat(fileName).st_mode & stat.S_IXUSR) != 0:
            return fileName
    return None


def gethostname() -> str:
    '''
    Returns current host name
    In fact, it's a wrapper for socket.gethostname()
    '''
    hostname = socket.gethostname()
    logger.info('Hostname: %s', hostname)
    return hostname


# Queing operations (to be executed before exit)


def register_for_delayed_deletion(filename: str, early_stage: bool = False) -> None:
    '''
    Adds a file to the wait-and-unlink list
    '''
    logger.debug('Added file %s to unlink on %s stage', filename, 'early' if early_stage else 'later')
    _unlink_files.append(types.RemovableFile(filename, early_stage))


def unlink_files(early_stage: bool = False) -> None:
    '''
    Removes all wait-and-unlink files
    '''
    logger.debug('Unlinking files on %s stage', 'early' if early_stage else 'later')
    files_to_unlink = list(filter(lambda x: x.early_stage == early_stage, _unlink_files))
    if files_to_unlink:
        logger.debug('Files to unlink: %s', files_to_unlink)
        # Wait 2 seconds before deleting anything on early and 5 on later stages
        time.sleep(1 + 2 * (1 + int(early_stage)))

        for f in files_to_unlink:
            try:
                os.unlink(f.path)
            except Exception as e:
                logger.debug('File %s not deleted: %s', f[0], e)

    # Remove all processed files from list
    _unlink_files[:] = list(set(_unlink_files) - set(files_to_unlink))


def add_task_to_wait(task: typing.Any, wait_subprocesses: bool = False) -> None:
    logger.debug(
        'Added task %s to wait %s',
        task,
        'with subprocesses' if wait_subprocesses else '',
    )
    _awaitable_tasks.append(types.AwaitableTask(task, wait_subprocesses))


def wait_for_tasks() -> None:
    logger.debug('Started to wait %s', _awaitable_tasks)
    for awaitable_task in _awaitable_tasks:
        logger.debug(
            'Waiting for task %s, subprocess wait: %s', awaitable_task.task, awaitable_task.wait_subprocesses
        )
        try:
            if hasattr(awaitable_task.task, 'join'):
                awaitable_task.task.join()
            elif hasattr(awaitable_task.task, 'wait'):
                awaitable_task.task.wait()
            # If wait for spanwed process (look for process with task pid) and we can look for them...
            if awaitable_task.wait_subprocesses and hasattr(awaitable_task.task, 'pid'):
                subprocesses: list['psutil.Process'] = list(
                    filter(
                        lambda x: x.ppid() == awaitable_task.task.pid,  # type x: psutil.Process
                        process_iter(attrs=('ppid',)),
                    )
                )
                logger.debug('Waiting for subprocesses... %s, %s', awaitable_task.task.pid, subprocesses)
                for i in subprocesses:
                    logger.debug('Found %s', i)
                    i.wait()
        except Exception as e:
            logger.error('Waiting for tasks to finish error: %s', e)

    # Empty the list
    _awaitable_tasks.clear()


def register_execute_before_exit(fnc: typing.Callable[[], None]) -> None:
    logger.debug('Registered exec before exit: %s', fnc)
    _execute_before_exit.append(fnc)


def execute_before_exit() -> None:
    logger.debug('Executing exec before exit: %s', _execute_before_exit)
    for fnc in _execute_before_exit:
        try:
            fnc()
        except Exception as e:
            logger.error('Error executing before exit: %s', e)
    _execute_before_exit.clear()


def verify_signature(script: bytes, signature: bytes) -> bool:
    '''
    Verifies with a public key from whom the data came that it was indeed
    signed by their private key
    param: public_key_loc Path to public key
    param: signature String signature to be verified
    return: Boolean. True if the signature is valid; False otherwise.
    '''
    public_key = serialization.load_pem_public_key(data=consts.PUBLIC_KEY, backend=default_backend())

    try:
        public_key.verify(  # type: ignore
            base64.b64decode(signature), script, padding.PKCS1v15(), hashes.SHA256()  # type: ignore
        )
    except Exception:  # InvalidSignature
        return False

    # If no exception, the script was fine...
    return True


def get_cacerts_file() -> typing.Optional[str]:
    # First, try certifi...

    # If environment contains CERTIFICATE_BUNDLE_PATH, use it
    if 'CERTIFICATE_BUNDLE_PATH' in os.environ:
        return os.environ['CERTIFICATE_BUNDLE_PATH']

    try:
        if os.path.exists(certifi.where()):
            return certifi.where()
    except Exception:
        pass

    logger.info('Certifi file does not exists: %s', certifi.where())

    # Check if "standard" paths are valid for linux systems
    if 'linux' in sys.platform:
        for path in (
            '/etc/pki/tls/certs/ca-bundle.crt',
            '/etc/ssl/certs/ca-certificates.crt',
            '/etc/ssl/ca-bundle.pem',
        ):
            if os.path.exists(path):
                logger.info('Found certifi path: %s', path)
                return path

    return None


def is_macos() -> bool:
    return 'darwin' in sys.platform

def is_linux() -> bool:
    return 'linux' in sys.platform

def is_windows() -> bool:
    return 'win' in sys.platform


# old compat names, to ensure compatibility with old code
# Will update server scripts to use new names as soon as previos 4.0 versions are not supported anymore
# (making them imcompatible with 3.x versions)
addTaskToWait = add_task_to_wait
# waitForTasks = wait_for_tasks  # Not used on scripts
saveTempFile = save_temp_file
readTempFile = read_temp_file
testServer = test_server
findApp = find_application
# getHostName = get_hostname  # Not used on scripts
addFileToUnlink = register_for_delayed_deletion
# unlinkFiles = unlink_files  # Not used on existing scripts
# isMac = is_mac_os  # Not used on scripts  # Not used on existing scripts
# getCaCertsFile = get_cacerts_file  # Not used on scripts
# verifySignature = verify_signature  # Not used on scripts
# execBeforeExit = exec_before_exit  # Not used on scripts
# addExecBeforeExit = register_execute_before_exit  # Not used on existing scripts



#!/usr/bin/env -S python3 -s
# -*- coding: utf-8 -*-
#
# Copyright (c) 2014-2024 Virtual Cable S.L.U.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without modification,
# are permitted provided that the following conditions are met:
#
#    * Redistributions of source code must retain the above copyright notice,
#      this list of conditions and the following disclaimer.
#    * Redistributions in binary form must reproduce the above copyright notice,
#      this list of conditions and the following disclaimer in the documentation
#      and/or other materials provided with the distribution.
#    * Neither the name of Virtual Cable S.L.U. nor the names of its contributors
#      may be used to endorse or promote products derived from this software
#      without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

# pyright: reportUnknownMemberType=false
# ,reportUnknownParameterType=false
# ,reportUnknownArgumentType=false
'''
Author: Adolfo Gómez, dkmaster at dkmon dot com
'''
import contextlib
import logging
import sys
import os
import platform
import time
import webbrowser
import threading
import urllib.parse
import typing

from uds.ui import QtCore, QtWidgets, QtGui, QSettings, Ui_MainWindow  # pyright: ignore
from uds.rest import RestApi

# Just to ensure there are available on runtime
from uds.tunnel import forward as tunnel_forwards  # pyright: ignore[reportUnusedImport]

from uds.log import logger, get_remote_log, init_remote_log
from uds import consts, exceptions, tools
from uds import VERSION


class UDSClient(QtWidgets.QMainWindow):
    ticket: str = ''
    scrambler: str = ''
    has_error = False
    animation_timer: typing.Optional[QtCore.QTimer] = None
    animation_value: int = 0
    animation_reversed: bool = False
    api: RestApi

    def __init__(self, api: RestApi, ticket: str, scrambler: str) -> None:
        QtWidgets.QMainWindow.__init__(self)
        self.api = api
        self.ticket = ticket
        self.scrambler = scrambler
        self.setWindowFlags(
            QtCore.Qt.WindowType.FramelessWindowHint | QtCore.Qt.WindowType.WindowStaysOnTopHint
        )

        self.ui = Ui_MainWindow()
        self.ui.setupUi(self)  # type: ignore

        self.ui.progressBar.setValue(0)
        self.ui.cancelButton.clicked.connect(self.on_cancel_pressed)

        self.ui.info.setText('Initializing...')

        screen_geometry = QtGui.QGuiApplication.primaryScreen().geometry()
        mysize = self.geometry()
        hpos = (screen_geometry.width() - mysize.width()) // 2
        vpos = (screen_geometry.height() - mysize.height() - mysize.height()) // 2
        self.move(hpos, vpos)

        self.animation_timer = QtCore.QTimer()
        self.animation_timer.timeout.connect(self.update_anim)
        # QtCore.QObject.connect(self.animTimer, QtCore.SIGNAL('timeout()'), self.updateAnim)

        self.start_animation()

    def close_window(self) -> None:
        self.close()

    def show_error(self, error: Exception) -> None:
        logger.error('Error: %s', error)
        self.stop_animation()
        # In fact, main window is hidden, so this is not visible... :)
        self.ui.info.setText('UDS Plugin Error')
        self.close_window()
        UDSClient.error_message('UDS Plugin Error', f'{error}')
        self.has_error = True

    def on_cancel_pressed(self) -> None:
        self.close()

    def update_anim(self) -> None:
        self.animation_value += 2
        if self.animation_value > 99:
            self.animation_reversed = not self.animation_reversed
            self.ui.progressBar.setInvertedAppearance(self.animation_reversed)
            self.animation_value = 0

        self.ui.progressBar.setValue(self.animation_value)

    def start_animation(self) -> None:
        self.ui.progressBar.setInvertedAppearance(False)
        self.animation_value = 0
        self.animation_reversed = False
        self.ui.progressBar.setInvertedAppearance(self.animation_reversed)
        if self.animation_timer:
            self.animation_timer.start(40)

    def stop_animation(self) -> None:
        self.ui.progressBar.setInvertedAppearance(False)
        if self.animation_timer:
            self.animation_timer.stop()

    def fetch_version(self) -> None:
        try:
            self.api.get_version()
        except exceptions.InvalidVersionException as e:
            UDSClient.error_message(
                'Upgrade required',
                f'UDS Client version {e.required_version} is required.\nA browser will be opened to download it.',
            )
            webbrowser.open(e.link)
            self.close_window()
            return
        except Exception as e:
            if logger.getEffectiveLevel() == logging.DEBUG:
                logger.exception('Get Version')
            self.show_error(e)
            self.close_window()
            return

        self.fetch_transport_data()

    def fetch_transport_data(self) -> None:
        try:
            script, params, log_data = self.api.get_script_and_parameters(self.ticket, self.scrambler)

            init_remote_log(log_data)  # Initialize for remote logging if requested by server

            self.stop_animation()

            if 'darwin' in sys.platform:
                self.showMinimized()

            self.close_window()

            # Execute UDS transport script, signed and checked
            vars = {
                '__builtins__': __builtins__,
                'parent': self,
                'sp': params,
            }
            exec(script, vars)

            self.process_waiting_tasks()

        except exceptions.RetryException as e:
            self.ui.info.setText(str(e) + ', retrying access...')
            # Retry operation in ten seconds
            QtCore.QTimer.singleShot(10000, self.fetch_transport_data)
        except Exception as e:
            # If debug is enabled, show exception
            if logger.getEffectiveLevel() == logging.DEBUG:
                logger.exception('Get Transport Data')
            self.show_error(e)
            process_remote_log(self.api)

    def process_waiting_tasks(self) -> None:
        """
        Process the waiting tasks in a separate thread.

        This way, the gui don't get blocked by the waiting tasks.

        Returns:
            None
        """

        threading.Thread(target=waiting_tasks_processor, args=(self.api,)).start()
        # And simply return

    def start(self) -> None:
        """
        Starts proccess by requesting version info
        """
        self.ui.info.setText('Initializing...')
        QtCore.QTimer.singleShot(100, self.fetch_version)  # Will make it async, not blocking the gui

    @staticmethod
    def warning_message(title: str, message: str, *, yes_no: bool = False) -> bool:
        buttons = (
            QtWidgets.QMessageBox.StandardButton.Yes | QtWidgets.QMessageBox.StandardButton.No
            if yes_no
            else QtWidgets.QMessageBox.StandardButton.Ok
        )
        return (
            QtWidgets.QMessageBox.warning(
                typing.cast(QtWidgets.QWidget, None),
                title,
                message,
                buttons,
            )
            == QtWidgets.QMessageBox.StandardButton.Yes  # If no yes_no, does not matter this comparison
        )

    @staticmethod
    def error_message(title: str, message: str, *, yes_no: bool = False) -> bool:
        buttons = (
            QtWidgets.QMessageBox.StandardButton.Yes | QtWidgets.QMessageBox.StandardButton.No
            if yes_no
            else QtWidgets.QMessageBox.StandardButton.Ok
        )
        return (
            QtWidgets.QMessageBox.critical(
                typing.cast(QtWidgets.QWidget, None),
                title,
                message,
                buttons,
            )
            == QtWidgets.QMessageBox.StandardButton.Yes  # If no yes_no, does not matter this comparison
        )

    @staticmethod
    @contextlib.contextmanager
    def settings(group: str) -> typing.Iterator[QSettings]:
        settings = QSettings()
        settings.beginGroup(group)
        yield settings
        settings.endGroup()

def process_remote_log(api: RestApi) -> None:
    # Process remote logging if requested
    try:
        log_ticket, log_data = get_remote_log()
        logger.debug('** Remote log data: %s, %s', log_ticket, len(log_data))
        if log_ticket != '' and len(log_data) > 0:
            logger.debug('** Sending log data: %s, %s', log_ticket, len(log_data))
            api.send_log(log_ticket, log_data[-65536:])  # Limit to 64K
    except Exception as e:  # pylint: disable=broad-exception-caught
        logger.error('** Error sending log data: %s', e)


def waiting_tasks_processor(api: RestApi) -> None:
    # Wait a bit before start processing ending sequence
    time.sleep(3)
    try:
        # Remove early stage files...
        tools.unlink_files(early_stage=True)
    except Exception as e:  # pylint: disable=broad-exception-caught
        logger.debug('Unlinking files on early stage: %s', e)

    # After running script, wait for stuff
    try:
        logger.debug('Wating for tasks to finish...')
        tools.wait_for_tasks()
    except Exception as e:  # pylint: disable=broad-exception-caught
        logger.debug('Watiting for tasks to finish: %s', e)

    try:
        logger.debug('Unlinking files')
        tools.unlink_files(early_stage=False)
    except Exception as e:  # pylint: disable=broad-exception-caught
        logger.debug('Unlinking files on later stage: %s', e)

    # Removing
    try:
        logger.debug('Executing threads before exit')
        tools.execute_before_exit()
    except Exception as e:  # pylint: disable=broad-exception-caught
        logger.debug('execBeforeExit: %s', e)

    logger.debug('endScript done')
    
    process_remote_log(api)


# Ask user to approve endpoint
def verify_host_approval(hostname: str) -> bool:
    with UDSClient.settings('endpoints') as settings:
        approved = bool(settings.value(hostname, False))

        errorString = '<p>The server <b>{}</b> must be approved:</p>'.format(hostname)
        errorString += '<p>Only approve UDS servers that you trust to avoid security issues.</p>'

        if not approved:
            if UDSClient.warning_message('ACCESS Warning', errorString, yes_no=True):
                settings.setValue(hostname, True)
                approved = True

    return approved


def ssl_certificate_validator(hostname: str, serial: str) -> bool:
    with UDSClient.settings('ssl') as settings:
        approved: bool = bool(settings.value(serial, False))
        if approved or UDSClient.warning_message(
            'SSL Warning',
            f'Could not check SSL certificate for {hostname}.\nDo you trust this host?',
            yes_no=True,
        ):
            approved = True
            settings.setValue(serial, True)

    return approved


# Used only if command line says so
def minimal(api: RestApi, ticket: str, scrambler: str) -> int:
    try:
        logger.info('Minimal Execution')
        logger.debug('Getting version')
        try:
            api.get_version()
        except exceptions.InvalidVersionException as e:
            UDSClient.error_message(
                'Upgrade required',
                'A newer connector version is required.\nA browser will be opened to download it.',
            )
            webbrowser.open(e.link)
            return 0
        logger.debug('Transport data')

        script, params, log_data = api.get_script_and_parameters(ticket, scrambler)

        init_remote_log(log_data)  # Initialize for remote logging if requested by server

        # A catch-all-calls class, to avoid errors on script becasue no parent
        class CatchAll:
            def __getattr__(self, name: str) -> 'CatchAll':
                return self

            def __call__(self, *args: typing.Any, **kwargs: typing.Any) -> 'CatchAll':
                return self

        vars = {
            '__builtins__': __builtins__,
            'parent': CatchAll(),
            'sp': params,
        }
        exec(script, vars)

        threading.Thread(target=waiting_tasks_processor, args=(api,)).start()

    except exceptions.RetryException as e:
        UDSClient.error_message(
            'Service not ready',
            '{}'.format('.\n'.join(str(e).split('.'))) + '\n\nPlease, retry again in a while.',
        )
    except Exception as e:  # pylint: disable=broad-exception-caught
        # logger.exception('Got exception on getTransportData')
        UDSClient.error_message('Error', f'{e}\n\nPlease, retry again in a while.')
    return 0


def parse_arguments(args: typing.List[str]) -> typing.Tuple[str, str, str, bool]:
    """
    Parses the command line arguments and returns a tuple containing the host, ticket, scrambler, and a flag indicating whether to use the minimal interface.

    Args:
        args (List[str]): The command line arguments. (including the program name as the first argument)

    Returns:
        Tuple[str, str, str, bool]: A tuple containing the host, ticket, scrambler, and a flag indicating whether to use the minimal interface.

    Raises:
        Exception: If the number of arguments is less than 2.
        IDSArgumentException: If the uds_url is '--test'.
        UDSMessageException: If the uds_url starts with 'uds://' and the DEBUG flag is False.
        UDSMessageException: If the uds_url does not start with 'udss://'.
    """
    if len(args) < 2:
        raise Exception()

    use_minimal_interface = False
    uds_url = args[1]

    if uds_url == '--minimal':
        use_minimal_interface = True
        uds_url = args[2]  # And get URI

    if uds_url == '--test':
        raise exceptions.ArgumentException('test')

    try:
        urlinfo = urllib.parse.urlparse(uds_url)
        ticket, scrambler = urlinfo.path.split('/')[1:3]
    except Exception:
        raise exceptions.MessageException('Invalid UDS URL')

    # Check if minimal interface is requested on the URL
    if 'minimal' in urllib.parse.parse_qs(urlinfo.query):
        use_minimal_interface = True

    if urlinfo.scheme == 'uds':
        if not consts.DEBUG:
            raise exceptions.MessageException(
                'UDS Client Version {} does not support HTTP protocol Anymore.'.format(VERSION)
            )
    elif urlinfo.scheme != 'udss':
        raise exceptions.MessageException('Not supported protocol')  # Just shows "about" dialog

    # If ticket length is not valid
    if len(ticket) != consts.TICKET_LENGTH:
        raise exceptions.MessageException(f'Invalid ticket: {ticket}')

    return (
        urlinfo.netloc,
        ticket,
        scrambler,
        use_minimal_interface,
    )


def main(args: typing.List[str]) -> int:
    app = QtWidgets.QApplication(sys.argv)
    logger.debug('Initializing connector for %s(%s)', sys.platform, platform.machine())

    logger.debug('Arguments: %s', args)
    # Set several info for settings
    QtCore.QCoreApplication.setOrganizationName('Virtual Cable S.L.U.')
    QtCore.QCoreApplication.setApplicationName('UDS Connector')

    if 'darwin' not in sys.platform:
        logger.debug('Mac OS *NOT* Detected')
        app.setStyle('plastique')
    else:
        logger.debug('Platform is Mac OS, adding homebrew well known paths')
        os.environ['PATH'] += ''.join(
            os.pathsep + i
            for i in (
                '/opt/homebrew/bin',
                '/usr/local/bin',
            )
        )
        logger.debug('Now path is %s', os.environ['PATH'])

    # First parameter must be url
    try:
        host, ticket, scrambler, _use_minimal_interface = parse_arguments(args)
    except exceptions.MessageException as e:
        logger.debug('Detected execution without valid URI, exiting: %s', e)
        UDSClient.error_message(
            f'UDS Client Version {VERSION}',
            f'{e}',
        )
        return 1
    except exceptions.ArgumentException as e:
        # Currently only test, return 0
        return 0
    except Exception:
        logger.debug('Detected execution without valid URI, exiting')
        UDSClient.error_message(
            'Notice',
            f'UDS Client Version {VERSION}',
        )

        return 1

    # Setup REST api and ssl certificate validator
    api = RestApi.api(
        host,
        on_invalid_certificate=ssl_certificate_validator,
    )

    try:
        logger.debug('Starting execution')

        # Approbe before going on
        if verify_host_approval(host) is False:
            raise exceptions.MessageException('Host {} was not approved'.format(host))

        win = UDSClient(api, ticket, scrambler)
        # Show and activate window, so it's on top
        win.show()
        win.activateWindow()

        win.start()

        exit_code = app.exec()
        logger.debug('Main execution finished correctly: %s', exit_code)

    except Exception as e:
        if not isinstance(e, exceptions.MessageException):
            logger.exception('Got an exception executing client:')
        else:
            logger.info('Message from error: %s', e)
        exit_code = 128
        UDSClient.error_message(
            'Error',
            f'Fatal error: {e}',
        )

    logger.debug('Exiting')
    return exit_code


if __name__ == "__main__":
    exit_code = main(sys.argv)
    # Set exit code
    sys.exit(exit_code)



# -*- coding: utf-8 -*-
#
# Copyright (c) 2023 Virtual Cable S.L.U.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without modification,
# are permitted provided that the following conditions are met:
#
#    * Redistributions of source code must retain the above copyright notice,
#      this list of conditions and the following disclaimer.
#    * Redistributions in binary form must reproduce the above copyright notice,
#      this list of conditions and the following disclaimer in the documentation
#      and/or other materials provided with the distribution.
#    * Neither the name of Virtual Cable S.L. nor the names of its contributors
#      may be used to endorse or promote products derived from this software
#      without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.



# -*- coding: utf-8 -*-
#
# Copyright (c) 2014-2024 Virtual Cable S.L.U.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without modification,
# are permitted provided that the following conditions are met:
#
#    * Redistributions of source code must retain the above copyright notice,
#      this list of conditions and the following disclaimer.
#    * Redistributions in binary form must reproduce the above copyright notice,
#      this list of conditions and the following disclaimer in the documentation
#      and/or other materials provided with the distribution.
#    * Neither the name of Virtual Cable S.L. nor the names of its contributors
#      may be used to endorse or promote products derived from this software
#      without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

'''
@author: Adolfo Gómez, dkmaster at dkmon dot com
'''
import io
import logging
import os
import os.path
import platform
import sys
import typing

from . import consts


# Local variables
_log_stream: typing.Optional[io.StringIO] = None  # Used to store log for remote log
_log_ticket = ''  # Used to store ticket for remote log


def _platform_debug_info(logger: 'logging.Logger') -> None:
    from . import ui  # To incldue qt version

    # Include as much as platform info as possible
    logger.debug('Platform info:')
    logger.debug('  UDSClient version: %s', consts.VERSION)
    logger.debug('  Platform: %s', platform.platform())
    logger.debug('  Node: %s', platform.node())
    logger.debug('  System: %s', platform.system())
    logger.debug('  Release: %s', platform.release())
    logger.debug('  Version: %s', platform.version())
    logger.debug('  Machine: %s', platform.machine())
    logger.debug('  Processor: %s', platform.processor())
    logger.debug('  Architecture: %s', platform.architecture())
    logger.debug('  Python version: %s', platform.python_version())
    logger.debug('  Python implementation: %s', platform.python_implementation())
    logger.debug('  Python compiler: %s', platform.python_compiler())
    logger.debug('  Python build: %s', platform.python_build())
    # Also environment variables and any useful info
    logger.debug('Qt framework: %s', ui.QT_VERSION)
    logger.debug('Log level set to DEBUG')
    logger.debug('Environment variables:')

    for k, v in os.environ.items():
        logger.debug('  %s=%s', k, v)

    # useful info for debugging
    logger.debug('Python path: %s', sys.path)
    logger.debug('Python executable: %s', sys.executable)
    logger.debug('Python version: %s', sys.version)
    logger.debug('Python version info: %s', sys.version_info)
    logger.debug('Python prefix: %s', sys.prefix)
    logger.debug('Python base prefix: %s', sys.base_prefix)
    logger.debug('Python executable: %s', sys.executable)
    logger.debug('Python argv: %s', sys.argv)
    logger.debug('Python modules path: %s', sys.path)
    logger.debug('Python modules importer cache path: %s', sys.path_importer_cache)
    logger.debug('Python modules hooks path: %s', sys.path_hooks)
    logger.debug('Python modules meta path: %s', sys.meta_path)


def _init() -> 'logging.Logger':
    try:
        logging.basicConfig(
            filename=consts.LOGFILE,
            filemode='a',
            format='%(levelname)s %(asctime)s %(message)s',
            level=consts.LOGLEVEL,
        )
    except Exception:
        logging.basicConfig(format='%(levelname)s %(asctime)s %(message)s', level=consts.LOGLEVEL)

    logger = logging.getLogger('udsclient')

    if consts.DEBUG:
        _platform_debug_info(logger)

    return logger


def init_remote_log(log_data: typing.Dict[str, typing.Any]) -> None:
    global _log_ticket, _log_stream

    try:
        if log_data.get('ticket') is not None:
            _log_ticket = log_data['ticket']

            _log_stream = io.StringIO()

            logger.debug('** Remote log requested : %s', log_data)

            log_level = log_data.get('level', logging.INFO)
            stream_handler = logging.StreamHandler(_log_stream)
            stream_handler.setLevel(logging.DEBUG)
            logger.debug('Setting log level as requested to %s', stream_handler.level)
            
            formatter = logging.Formatter('%(levelname)s %(asctime)s %(message)s')
            stream_handler.setFormatter(formatter)
            logger.addHandler(stream_handler)
            
            # Dissable logging to parent handlers, so we only log to the stream handler
            # if local debug is not enabled
            if not consts.DEBUG:
                logger.propagate = False
            
            # Ensure log level is set to the requested one
            logger.setLevel(log_level)
            
            logger.debug('Handlers: %s', logger.handlers)
            
            # Repeat platform info for remote (will be duplicated locally if debug is enabled, but that's ok)
            _platform_debug_info(logger)

    except Exception as e:
        logger.error('Error setting log level as requested to %s: %s', log_data.get('level'), e)


def get_remote_log() -> typing.Tuple[str, str]:
    if _log_stream is None:
        return '', ''

    return _log_ticket, _log_stream.getvalue()


# Initialize logger
logger: 'logging.Logger' = _init()
